import asyncio
import enum
import io, os, torch
import json
import queue
import threading
import uuid
from threading import Thread

import numpy as np
from aiges.core.types import *

try:
    from aiges_embed import ResponseData, Response, DataListNode, DataListCls, SessionCreateResponse, callback  # c++
except:
    from aiges.dto import Response, ResponseData, DataListNode, DataListCls, SessionCreateResponse, callback

from aiges.sdk import WrapperBase
from aiges.utils.log import log, getFileLogger
import time
from belt_nlp.bert_classifier_with_pooling import BertClassifierWithPooling


class RequestMode(enum.Enum):
    ONCE = 'once'
    ONCE_ASYNC = 'once_async'
    STREAM = 'stream'
    STOP = 'stop'


def random_uuid() -> str:
    return str(uuid.uuid4().hex)


class RequestInfo:
    def __init__(self, sid: str, params: dict, user_tag: str = ""):
        self.handle = str(uuid.uuid4())
        self.sid = sid
        self.user_tag = user_tag
        self.params = params
        self.requests = []
        self.image_slices = []
        self.text_slices = []
        self.id2label = {}


class PromptInferenceInfo:
    def __init__(self, wrapper,
                 thread_id: str,
                 mode: RequestMode,
                 prompt: str,
                 requestInfo: RequestInfo,
                 result_q: queue.Queue = None,
                 functions: list = None):
        self.wrapper = wrapper
        self.requestInfo = requestInfo
        self.thread_id = thread_id
        self.mode = mode
        self.prompt = {}
        self.functions = functions
        self.request_id = random_uuid()
        self.result_q = result_q


def resp_content(status, index, text):
    resd = ResponseData()
    resd.key = "content"
    resd.setDataType(DataText)
    resd.status = status
    resd.setData(json.dumps(text).encode("utf-8"))
    return resd


def resp_usage(status, prompt_token, completion_token):
    resp_json = {"completion_tokens": completion_token, "prompt_tokens": prompt_token, "question_tokens": 4,
                 "total_tokens": prompt_token + completion_token}
    resd = ResponseData()
    resd.key = "usage"
    resd.setDataType(DataText)
    resd.status = status
    resd.setData(json.dumps(resp_json).encode("utf-8"))
    return resd


class ThreadPool:
    def __init__(self, num_threads, wrapper):
        self.num_threads = num_threads
        self.threads = {}
        self.task_queues = {}
        self.lock = threading.Lock()
        self.wrapper = wrapper
        self.max_batch_size = 8
        self.timeout = 0.01
        import logging
        self.filelogger = getFileLogger(level=logging.INFO)

        for i in range(num_threads):
            process_id = os.getpid()
            thread_id = "process-{}-thread-{}".format(str(process_id), str(i))
            task_queue = queue.Queue()
            self.threads[thread_id] = threading.Thread(target=self.worker, args=(thread_id, task_queue))
            self.task_queues[thread_id] = task_queue

        for thread in self.threads.values():
            thread.start()

    async def bert_loop(self, task_queue):
        self.filelogger.debug(f"bert_loop enter")
        running = True
        self.filelogger.info(f"bert_loop started")

    def worker(self, thread_id, task_queue):
        # 创建并运行事件循环
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(self.bert_loop(task_queue))

    def alloc_min_thread(self) -> str:
        with self.lock:
            min_thread_id = min(self.threads, key=lambda thread_id: self.task_queues[thread_id].qsize())
        return min_thread_id

    def put_task(self, thread_id, task):
        with self.lock:
            self.task_queues[thread_id].put(task)

    def wait_completion(self):
        for task_queue in self.task_queues.values():
            task_queue.put(None)

        for thread in self.threads.values():
            thread.join()


class WrapperFactory():
    def __init__(self, pretrained_name, base_model_path, request_map, lock):
        self.pretrained_name = pretrained_name
        self.base_model_path = base_model_path
        self.model = None
        self.tokenizer = None
        self.processor = None
        import logging
        self.filelogger = getFileLogger(level=logging.INFO)
        self.request_map = request_map
        self.lock = lock

    def load_model_tokenizer(self):
        pass

    def infer(self, params, reqData: DataListCls):
        pass

    def base_predict(self, reqData: RequestInfo):
        pass

    def trained_predict(self, reqData: DataListCls, params, patch_id):
        pass

    def get_payload_messages(self, reqData: DataListCls):
        pass

    @classmethod
    def create_wrapper_cls(cls, pretrain_name, base_model_path, requestMap, lock):
        from ailab.atp_finetuner.constant import Model
        BeltModel = [Model.bert_base_chinese, Model.chinese_roberta_wwm_ext]
        if pretrain_name in BeltModel:
            return Belt(pretrain_name, base_model_path, requestMap, lock)


class Belt(WrapperFactory):
    def load_model_tokenizer(self):
        model = BertClassifierWithPooling.load(self.base_model_path)
        self.model = model
        id2lable_file = os.path.join(self.base_model_path, 'id2label.json')
        with open(id2lable_file, 'r') as file:
            tag_data = json.load(file)
        self.tag_data = tag_data

    def infer(self, params, reqData: DataListCls):
        self.filelogger.info(f'once_infer params {params}')
        ori_input = self.get_payload_messages(reqData)
        res_tag = self.base_predict(ori_input)
        self.filelogger.info(f'once_infer result: {res_tag}')
        return res_tag

    def base_predict(self, req_context: str):
        predictions = self.model.predict_scores(req_context)
        predicted_classes = np.array(predictions.cpu)
        self.filelogger.info(f'predicted_classes: {predicted_classes}')

        category_scores = {}
        for i, p in enumerate(predictions):
            for idx, prob in enumerate(p):
                category = self.tag_data.get(str(idx), None)
                if category:
                    category_scores[category] = float(prob)
        res_tag = {
            "category_scores": category_scores
        }

        # predicted_index = np.argmax(predictions.cpu().numpy(), axis=1)[0]
        self.filelogger.info(f'res_tag: {res_tag}')
        return res_tag

    def get_payload_messages(self, reqData: DataListCls):
        self.filelogger.info(f'get_payload_messages reqData: {reqData}')
        messagesStr = reqData.get('messages').data.decode('utf-8')
        # 校验格式
        try:
            messages = json.loads(messagesStr)
            message = ""
            for item in messages['messages']:
                if item['role'] == "user":
                    message = item['content']
        except Exception as e:
            import traceback
            traceback.print_exc()
            return None
        self.filelogger.info(f'message: {message}')
        batch = []
        batch.append(message)
        return batch


class Wrapper(WrapperBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        import logging, threading
        self.filelogger = getFileLogger(level=logging.INFO)
        self.patch_id = {}
        self.lock = threading.Lock()
        # self.patch_id_lock = threading.Lock()
        self.request_map = {}
        self.request_map_lock = threading.Lock()
        self.thread_pool_size = 1  # Multithreading may not be supported, vllm
        self.thread_pool = None
        self.wrapper_cls = None

    def wrapperInit(self, config: {}) -> int:
        self.filelogger.info("Initializing ...")
        self.pretrained_name = os.environ.get("PRETRAINED_MODEL_NAME")
        self.base_model_path = os.environ.get("FULL_MODEL_PATH")

        self.wrapper_cls = WrapperFactory.create_wrapper_cls(self.pretrained_name, self.base_model_path,
                                                             self.request_map, self.request_map_lock)
        self.wrapper_cls.load_model_tokenizer()
        self.filelogger.info(f'wrapper init success, create thread:{self.thread_pool_size}')
        return 0

    def wrapperOnceExec(self, params: {}, reqData: DataListCls, usrTag: str = "", persId: int = 0) -> Response:
        self.filelogger.info(f'input: {input}')
        res = Response()
        self.lock.acquire()

        try:
            res_tag = self.wrapper_cls.infer(params, reqData)
        except Exception as e:
            self.filelogger.debug("Infer Error:  %s" % str(e))
            return res.response_err(400)
        finally:
            self.lock.release()

        resData = ResponseData()
        resData.setData(json.dumps(res_tag, ensure_ascii=False).encode("utf-8"))
        resData.setDataType(DataText)
        resData.status = DataOnce
        resData.key = "score"
        res.list = [resData]
        return res

    def wrapperFini(self) -> int:
        self.thread_pool.wait_completion()
        return 0

    def wrapperError(cls, ret: int) -> str:
        if ret == 100:
            return "user error defined here"
        if ret == 400:
            return "请求体解析异常"
        return ""

    '''
        此函数保留测试用，不可删除
    '''

    def wrapperTestFunc(cls, data: [], respData: []):
        pass

    def wrapperCreate(self, params: {}, sid: str, persId: int = 0, usrTag: str = "") -> SessionCreateResponse:
        requestInfo = RequestInfo(sid, params, usrTag)
        self.request_map_lock.acquire()
        self.request_map[requestInfo.handle] = requestInfo
        self.request_map_lock.release()

        s = SessionCreateResponse()
        s.handle = requestInfo.handle
        s.error_code = 0
        self.filelogger.debug(f'success wrapperCreate handle {requestInfo.handle}')
        self.wrapper_cls.wrapperCreate(requestInfo)
        return s

    def wrapperWrite(self, handle: str, req: DataListCls) -> int:
        self.filelogger.debug(f'start wrapperWrite handle {handle}')

        self.request_map_lock.acquire()
        requestInfo = self.request_map[handle]
        if requestInfo == None:
            self.filelogger.error("can't get this handle:" % handle)
            return -1
        self.request_map_lock.release()

        thread_id = self.thread_pool.alloc_min_thread()
        prompt = None
        inferenceInfo = PromptInferenceInfo(self, thread_id, RequestMode.STREAM, prompt, requestInfo, functions=None)
        self.thread_pool.put_task(thread_id, inferenceInfo)
        self.request_map_lock.acquire()
        self.request_map[handle].requests.append(inferenceInfo.request_id)
        self.request_map_lock.release()
        self.filelogger.debug(
            f'success wrapperWrite handle:{handle},thread_id:{thread_id},request_id:{inferenceInfo.request_id}')
        return 0
        # return self.wrapper_cls.wrapperWrite(requestInfo, req)

    def wrapperDestroy(self, handle: str) -> int:
        self.filelogger.debug(f'start wrapperDestroy {handle}')
        self.request_map_lock.acquire()
        requestInfo = self.request_map[handle]
        if requestInfo == None:
            self.filelogger.error("can't get this handle:" % handle)
            self.request_map_lock.release()
            return -1
        del self.request_map[handle]
        self.request_map_lock.release()
        self.filelogger.debug(f'success wrapperDestroy {handle}')
        return 0